/*

  Copyright (C) 2001 StrmnNrmn

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// TODO: Add a CriticalSection to stop multiple threads using the 
// Unique objects simultaneously

#include "stdafx.h"
#include "DaedalusGraphics.h"
#include "GraphicsPlugin_Spec.h"

#include "Debug\DBGConsole.h"

#include "Graphics\D3DRender.h"
#include "Graphics\GraphicsContext.h"
#include "Graphics\TextureCache.h"
#include "Graphics\DLParser.h"

#include "Interface\IniFile.h"

#include "Utility\Profiler.h"

#define DAEDALUSGRAPHICS_API __declspec(dllexport)

#define DAEDALUS_GFX_PLUGIN_VERSION		0x0001


CHAR g_szDaedalusExeDir[ MAX_PATH + 1 ] = "";
CHAR g_szDaedalusIniFile[ MAX_PATH + 1 ] = "";
CHAR g_szDaedalusName[] = "Daedalus";

BOOL g_bDisplayNewCombiner = FALSE;
BOOL g_bTrapExceptions = FALSE;

DWORD g_dwDaedalusVersion = 0;

u8 * g_pu8RamBase = NULL;
u32 * g_pu32RamBase = NULL;
s8 * g_ps8RamBase = NULL;
u8 * g_pu8SpMemBase = NULL;

RomInfo g_ROM;

DWORD g_dwRamSize = 8*1024*1024;

#define OS_TV_PAL		0
#define OS_TV_NTSC		1
#define OS_TV_MPAL		2

GFX_INFO g_GraphicsInfo;

BOOL g_bDropAllTextures = FALSE;
BOOL g_bDumpNextScreen = FALSE;

CCritSect g_CritialSection;

HINSTANCE g_hModule = NULL;

static void InitialisePaths();

//*****************************************************************************
//
//*****************************************************************************
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	HRESULT hr;

    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:

			DisableThreadLibraryCalls( (HINSTANCE)hModule );

			g_hModule = (HINSTANCE)hModule;
			InitialisePaths();

			hr = CProfiler::Create();
			if ( FAILED(hr) )
				return FALSE;

			hr = CGraphicsContext::Create();
			if ( FAILED(hr) )
				return FALSE;

			hr = D3DRender::Create();
			if ( FAILED(hr) )
				return FALSE;
			
			hr = CTextureCache::Create();
			if (FAILED(hr))
				return hr;

			break;

		// These should not be called because of DisableThreadLibraryCalls
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
			break;


		case DLL_PROCESS_DETACH:

			CTextureCache::Destroy();
			D3DRender::Destroy();
			CGraphicsContext::Destroy();
			CProfiler::Destroy();

			break;
    }

    return TRUE;
}

//*****************************************************************************
// Construct g_szDaedalusExeDir and g_szDaedalusIniFile
//*****************************************************************************
void InitialisePaths()
{
	HRESULT hr;

	GetModuleFileName( g_hModule, g_szDaedalusExeDir, sizeof(g_szDaedalusExeDir) );
	PathRemoveFileSpec( g_szDaedalusExeDir );

	// Try to find the ini file
	PathCombine(g_szDaedalusIniFile, g_szDaedalusExeDir, TEXT("daedalus.ini"));	
	if ( !PathFileExists( g_szDaedalusIniFile ))
	{
		// Try parent directory
		lstrcpy( g_szDaedalusIniFile, g_szDaedalusExeDir );
		PathRemoveFileSpec( g_szDaedalusIniFile );
		PathAppend( g_szDaedalusIniFile, TEXT("daedalus.ini") );
		if ( !PathFileExists( g_szDaedalusIniFile ))
		{
			OutputDebugString("Couldn't find Daedalus.ini!\n");
		}
	}

	//
	// Load the inifile
	//
	hr = CIniFile::Create();
	if (FAILED(hr))
		return;

	BOOL bResult = CIniFile::Get()->OpenIniFile( g_szDaedalusIniFile );
	if (!bResult)
	{
		MessageBox(NULL, "Unable to read IniFile from disk", g_szDaedalusName, MB_OK);
		//return;
	}
	
}


/******************************************************************
  Function: GetDllInfo
  Purpose:  This function allows the emulator to gather information
            about the dll by filling in the PluginInfo structure.
  input:    a pointer to a PLUGIN_INFO stucture that needs to be
            filled by the function. (see def above)
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void GetDllInfo ( PLUGIN_INFO * PluginInfo )
{
	PluginInfo->Version = 0x102;
	PluginInfo->Type = PLUGIN_TYPE_GFX;
	strcpy( PluginInfo->Name, "DaedalusGraphics" );
	PluginInfo->NormalMemory = TRUE;
	PluginInfo->MemoryBswaped = TRUE;
}

/******************************************************************
  Function: InitiateGFX
  Purpose:  This function is called when the DLL is started to give
            information from the emulator that the n64 graphics
			uses. This is not called from the emulation thread.
  Input:    Gfx_Info is passed to this function which is defined
            above.
  Output:   TRUE on success
            FALSE on failure to initialise
             
  ** note on interrupts **:
  To generate an interrupt set the appropriate bit in MI_INTR_REG
  and then call the function CheckInterrupts to tell the emulator
  that there is a waiting interrupt.
*******************************************************************/ 
DAEDALUSGRAPHICS_API BOOL InitiateGFX (GFX_INFO Gfx_Info)
{

	g_GraphicsInfo = Gfx_Info;

	g_pu8RamBase = (u8*)Gfx_Info.RDRAM;
	g_ps8RamBase = (s8*)Gfx_Info.RDRAM;
	g_pu32RamBase = (u32*)Gfx_Info.RDRAM;
	g_pu8SpMemBase = (u8*)g_GraphicsInfo.DMEM;

	strcpy( g_ROM.settings.szGameName, "Wibble" );
	g_ROM.settings.ucode = 0;
	g_ROM.settings.bDisableTextureCRC = FALSE;
	g_ROM.nTvType = OS_TV_NTSC;

	return TRUE;
}

/******************************************************************
  Function: CloseDLL
  Purpose:  This function is called when the emulator is closing
            down allowing the dll to de-initialise.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void CloseDLL (void)
{
	// Do nothing - do processing in DLL_PROCESS_DETACH
	g_CritialSection.Lock();
	CGraphicsContext::Get()->Cleanup3DEnvironment();
	g_CritialSection.Unlock();
}

/******************************************************************
  Function: ChangeWindow
  Purpose:  to change the window between fullscreen and window 
            mode. If the window was in fullscreen this should 
			change the screen to window mode and vice vesa.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void ChangeWindow (void)
{
	g_CritialSection.Lock();
	CGraphicsContext::Get()->ToggleFullscreen();
	g_CritialSection.Unlock();
}

/******************************************************************
  Function: DllAbout
  Purpose:  This function is optional function that is provided
            to give further information about the DLL.
  input:    a handle to the window that calls this function
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void DllAbout ( HWND hParent )
{
	MessageBox( hParent, "DeadalusGraphics\nCopyright (c) 2001 StrmnNrmn", "DaedalusGraphics", MB_OK);
}

/******************************************************************
  Function: DllConfig
  Purpose:  This function is optional function that is provided
            to allow the user to configure the dll
  input:    a handle to the window that calls this function
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void DllConfig ( HWND hParent )
{
	g_CritialSection.Lock();
	CGraphicsContext::Get()->OnChangeDevice();
	g_CritialSection.Unlock();
}

/******************************************************************
  Function: DllTest
  Purpose:  This function is optional function that is provided
            to allow the user to test the dll
  input:    a handle to the window that calls this function
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void DllTest ( HWND hParent )
{
	// TODO
	g_CritialSection.Lock();
	g_CritialSection.Unlock();
}

/******************************************************************
  Function: DrawScreen
  Purpose:  This function is called when the emulator receives a
            WM_PAINT message. This allows the gfx to fit in when
			it is being used in the desktop.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void DrawScreen (void)
{
	// TODO
	g_CritialSection.Lock();
	if ( CGraphicsContext::IsAvailable() )
	{
		CGraphicsContext::Get()->UpdateFrame();
	}
	g_CritialSection.Unlock();
}


/******************************************************************
  Function: MoveScreen
  Purpose:  This function is called in response to the emulator
            receiving a WM_MOVE passing the xpos and ypos passed
			from that message.
  input:    xpos - the x-coordinate of the upper-left corner of the
            client area of the window.
			ypos - y-coordinate of the upper-left corner of the
			client area of the window. 
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void MoveScreen (int xpos, int ypos)
{
	// TODO
	g_CritialSection.Lock();
	g_CritialSection.Unlock();
}

/******************************************************************
  Function: ProcessDList
  Purpose:  This function is called when there is a Dlist to be
            processed.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void ProcessDList(void)
{
	g_CritialSection.Lock();


	if (g_bDropAllTextures)
	{
		CTextureCache::Get()->DropTextures();
		g_bDropAllTextures = FALSE;
	}

	DLParser_Process();

	if ( g_bDumpNextScreen )
	{
		CGraphicsContext::Get()->DumpScreenShot();
		g_bDumpNextScreen = FALSE;
	}


//	CGraphicsContext::Get()->UpdateFrame();


	g_CritialSection.Unlock();
}

/******************************************************************
  Function: RomClosed
  Purpose:  This function is called when a rom is closed.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void RomClosed (void)
{
	g_CritialSection.Lock();

	CTextureCache::Get()->DropTextures();

	g_CritialSection.Unlock();
}

/******************************************************************
  Function: RomOpen
  Purpose:  This function is called when a rom is open. (from the 
            emulation thread)
  input:    none
  output:   none
*******************************************************************/ 
typedef struct
{
	s8	nCountryID;
	LPCTSTR szName;
	u32 nTvType;
} CountryIDInfo;

const CountryIDInfo g_CountryCodeInfo[] = 
{
	{ 0,   TEXT("0"), OS_TV_NTSC },
	{ '7', TEXT("Beta"), OS_TV_NTSC },
	{ 'A', TEXT("NTSC"), OS_TV_NTSC },
	{ 'D', TEXT("Germany"), OS_TV_PAL },
	{ 'E', TEXT("USA"), OS_TV_NTSC },
	{ 'F', TEXT("France"), OS_TV_PAL },
	{ 'I', TEXT("Italy"), OS_TV_PAL },
	{ 'J', TEXT("Japan"), OS_TV_NTSC },
	{ 'P', TEXT("Europe"), OS_TV_PAL },
	{ 'S', TEXT("Spain"), OS_TV_PAL },
	{ 'U', TEXT("Australia"), OS_TV_PAL },
	{ 'X', TEXT("PAL"), OS_TV_PAL },
	{ 'Y', TEXT("PAL"), OS_TV_PAL },
	{ -1,  NULL, OS_TV_NTSC }
};

DAEDALUSGRAPHICS_API void RomOpen (void)
{
	HRESULT hr;

	g_CritialSection.Lock();
	
	memcpy( &g_ROM.rh, g_GraphicsInfo.HEADER, sizeof(g_ROM.rh) );

	RomID id( g_ROM.rh.CRC1, g_ROM.rh.CRC2, g_ROM.rh.CountryID );

	if ( !CIniFile::Get()->GetSettings( id, &g_ROM.settings ) )
	{
		// Default?
	}


	g_ROM.nTvType = OS_TV_NTSC;
	for ( u32 i = 0; g_CountryCodeInfo[i].szName != NULL; i++)
	{
		if (g_CountryCodeInfo[i].nCountryID == g_ROM.rh.CountryID)
		{
			g_ROM.nTvType = g_CountryCodeInfo[i].nTvType;
			break;
		}
	}

	hr = CGraphicsContext::Get()->Initialise( g_GraphicsInfo.hWnd, g_GraphicsInfo.hStatusBar, 640, 480, TRUE );
	if (SUCCEEDED(hr)) 
	{
		hr = DLParser_Init();
		if (SUCCEEDED(hr)) 
		{
			//
		}
	}

	g_CritialSection.Unlock();
}

/******************************************************************
  Function: UpdateScreen
  Purpose:  This function is called in response to a vsync of the
            screen were the VI bit in MI_INTR_REG has already been
			set
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void UpdateScreen (void)
{
	g_CritialSection.Lock();

	static u32 origin = 0;

	if ( *g_GraphicsInfo.xVI_ORIGIN_REG != origin )
	{
		CGraphicsContext::Get()->UpdateFrame();
		origin = *g_GraphicsInfo.xVI_ORIGIN_REG;
	}

	g_CritialSection.Unlock();
}

/******************************************************************
  Function: ViStatusChanged
  Purpose:  This function is called to notify the dll that the
            ViStatus registers value has been changed.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void ViStatusChanged (void)
{
	// TODO
	g_CritialSection.Lock();
	g_CritialSection.Unlock();
}

/******************************************************************
  Function: ViWidthChanged
  Purpose:  This function is called to notify the dll that the
            ViWidth registers value has been changed.
  input:    none
  output:   none
*******************************************************************/ 

DAEDALUSGRAPHICS_API void ViWidthChanged (void)
{
	// We ignore this - we just recalculate the width/height when we start processing a dlist
}



//*****************************************************************************
// Execute an arbitary command (e.g. "Take ScreenShot", "Dump DList")
// pszCommand: [in] Command name
// ppResult: [in/out] Optional auxillary data
// Returns standard HRESULT (S_OK, E_FAIL, E_OUTOFMEMORY etc)
//*****************************************************************************

DAEDALUSGRAPHICS_API HRESULT ExecuteCommand( const CHAR * pszCommand, void * pResult )
{
	HRESULT hr = E_NOTIMPL;

	g_CritialSection.Lock();

	if ( lstrcmpi( pszCommand, DAEDALUS_GFX_SETDAEDALUSVERSION ) == 0 )
	{
		g_dwDaedalusVersion = (DWORD)pResult;
		hr = S_OK;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_GETDAEDALUSPLUGINVERSION ) == 0 )
	{
		*(u32*)pResult = DAEDALUS_GFX_PLUGIN_VERSION;
		hr = S_OK;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_CAPTURESCREEN ) == 0 )
	{
		g_bDumpNextScreen = TRUE;
		hr = S_OK;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_DROPTEXTURES ) == 0 )
	{
		g_bDropAllTextures = TRUE;
		hr = S_OK;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_DUMPTEXTURES ) == 0 )
	{
		CTextureCache::Get()->SetDumpTextures( TRUE );
		g_bDropAllTextures = TRUE;		// Drop textures so that we can re-create
		hr = S_OK;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_NODUMPTEXTURES ) == 0 )
	{
		CTextureCache::Get()->SetDumpTextures( FALSE );
		hr = S_OK;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_USENEWCOMBINER ) == 0)
	{
		g_bDisplayNewCombiner = (BOOL)pResult;
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_DUMPDL ) == 0 )
	{
		DLParser_DumpNextDisplayList();
	}
	else if ( lstrcmpi( pszCommand, DAEDALUS_GFX_SET_DEBUG ) == 0 )
	{
		OutputDebugString( "Attaching debug console" );
		CDebugConsole::Attach( static_cast< CDebugConsole * >( pResult ) );
	}

	g_CritialSection.Unlock();
	return hr;
}







/*
void IMainWindow::SizeForDisplay()
{
	RECT rcScreen;
	RECT rcStatus;

	// This might not be the best flag to switch on - ]
	// maybe we could check g_bD3DXReady?
	if (g_bCPURunning)
	{
		SetRect(&rcScreen, 0,0, g_dwDisplayWidth, g_dwDisplayHeight);

		// Calculate window size to give desired window size...
		AdjustWindowRectEx(&rcScreen, WS_OVERLAPPEDWINDOW, TRUE, 0);

		// Add on enough space for the status bar
		GetClientRect(m_hWndStatus, &rcStatus);
		rcScreen.bottom += (rcStatus.bottom - rcStatus.top);

		SetWindowPos(m_hWnd, 0,
			rcScreen.left, rcScreen.top,
			rcScreen.right-rcScreen.left, rcScreen.bottom-rcScreen.top,
			SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOZORDER);
	}
	else
	{
		rcScreen = g_rcMainWindow;

		SetWindowPos(m_hWnd, 0,
			rcScreen.left, rcScreen.top,
			rcScreen.right-rcScreen.left, rcScreen.bottom-rcScreen.top,
			SWP_NOACTIVATE|SWP_NOZORDER);
	}
}
*/