/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __TEXTURECACHE_H__
#define __TEXTURECACHE_H__

#include "Texture.h"
#include "Utility\CritSect.h"
#include "Utility\Unique.h"

struct TextureInfo
{
	u32		TmemAddress;
	u32		TmemPalAddress;

	u32		Address;		// Corresponds to dwAddress in Tile
	u32		Format;			// e.g. RGBA, IA
	u32		Size;			// e.g. 16bpp
	s32		Left;
	s32		Top;
	u32		Width;
	u32		Height;
	u32		Pitch;
	u32		TLutFmt;

	bool	Swapped;		// Are odd lines word swapped?

	// or bool operator==(,,) ?
	bool IsEqual(const TextureInfo & ti) const
	{
		return ( memcmp(this, &ti, sizeof(TextureInfo)) == 0 );
	}

};

class CTextureCache : public CCritSect, public CUnique< CTextureCache >
{
	public:
		virtual ~CTextureCache() {};

		virtual void PurgeOldTextures() = 0;
		virtual void DropTextures() = 0;

		virtual void SetDumpTextures( BOOL bDumpTextures ) = 0;
		virtual BOOL GetDumpTextures( ) const = 0;

#ifdef MIRRORED_TEXTURE_SUPPORT
		virtual CTexture * GetTexture(TextureInfo * pti, BOOL bMirrored) = 0;
#else
		virtual CTexture * GetTexture( TextureInfo * pti ) = 0;
#endif

		virtual HRESULT InvalidateDeviceObjects() = 0;
};

#endif	// __TEXTURECACHE_H__
