/*
Copyright (C) 2001 CyRUS64 (http://www.boob.co.uk)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "IniFile.h"
#ifndef DAEDALUS_GFX_PLUGIN
#include "Core\ROM.h"
#include "RomDB.h"
#endif

//*****************************************************************************
//
//*****************************************************************************
class IIniFile : public CIniFile
{
	public:
		IIniFile();
		virtual ~IIniFile();

		//
		// CIniFile implementation
		//
		bool			OpenIniFile( const CHAR * filename );
		void			Commit();												// (STRMNNRMN - Write ini back out to disk?)

		bool			GetSettings( const RomID & id, RomSettings * p_settings ) const;
		void			SetSettings( const RomID & id, const RomSettings & settings );

	private:

		//
		// Utility functions
		//
		int				FindEntry( const RomID & id ) const; 		// Find entry in list
		void			AddEntry( const RomID & id, const RomSettings & settings );

		std::istream &	getline( std::istream & is, char *str );
		void			OutputSectionDetails(DWORD i, FILE * fh);

	private:

		struct section
		{
			RomID				mRomID;
			RomSettings			mSettings;
			bool				bOutput;			// (STRMNNRMN - Has it been written to disk yet?)
		};

		std::vector<section>	mSections;

		bool					mDirty;				// (STRMNNRMN - Changed since read from disk?)
		TCHAR					mFileName[ MAX_PATH + 1 ];
};



//*****************************************************************************
// Singleton creator
//*****************************************************************************
HRESULT CUnique< CIniFile >::Create()
{
	_ASSERTE(m_pInstance == NULL);
	
	m_pInstance = new IIniFile();
	if (!m_pInstance)
		return E_OUTOFMEMORY;

	return S_OK;
};


//*****************************************************************************
// Constructor
//*****************************************************************************
IIniFile::IIniFile()
{
	mSections.clear();
	mDirty = false;
	mFileName[ 0 ] = '\0';
}

//*****************************************************************************
//
//*****************************************************************************
IIniFile::~IIniFile()
{
	if ( mDirty )
	{
		Commit();
	}
}

//*****************************************************************************
//
//*****************************************************************************
static char * left(char * src, int nchars)
{
	static char dst[300];			// BUGFIX (STRMNNRM)
	strncpy(dst,src,nchars);
	dst[nchars]=0;
	return dst;
}

//*****************************************************************************
//
//*****************************************************************************
static char * right(char *src, int nchars)
{
	static char dst[300];			// BUGFIX (STRMNNRM)
	strncpy(dst, src + strlen(src) - nchars, nchars);
	dst[nchars]=0;
	return dst;
}

//*****************************************************************************
// (STRMNNRMN - Strip spaces from end of names)
//*****************************************************************************
static char * tidy(char * s)
{
	char * p = s + lstrlen(s);

	p--;
	while (p >= s && *p == ' ')
	{
		*p = 0;
		p--;
	}
	return s;

}

//*****************************************************************************
//
//*****************************************************************************
static RomID	RomIDFromString( const CHAR * str )
{
	u32 crc1, crc2, country;
	sscanf( str, "%08x%08x-%02x", &crc1, &crc2, &country );
	return RomID( crc1, crc2, (u8)country );
}


//*****************************************************************************
//
//*****************************************************************************
bool IIniFile::OpenIniFile( const CHAR * filename )
{
	std::ifstream inifile;
	char readinfo[100];
	char trim[]="{}"; //remove first and last character

	//
	// Remember the filename
	//
	strcpy( mFileName, filename );
	
	inifile.open(mFileName);
	
	if (inifile.fail())
		return false;

	while (getline(inifile,readinfo)/*&&sectionno<999*/)
	{
		tidy(readinfo);			// STRMNNRMN - Strip spaces from end of lines

		// STRMNNRMN - comments
		if (readinfo[0] == '/')
			continue;

		if (!lstrcmpi(readinfo,"")==0)
		{
			if (readinfo[0] == '{') //if a section heading
			{
				section newsection;

				StrTrim(readinfo,trim);

				newsection.mRomID = RomIDFromString( readinfo );
				//strcpy(newsection.crccheck, readinfo);
				newsection.bOutput = false;
				ROM_ResetSettings( &newsection.mSettings );

				mSections.push_back(newsection);
				
			}
			else
			{		
				int sectionno = mSections.size() - 1;

				if (lstrcmpi(left(readinfo,5), "UCode")==0)
					mSections[sectionno].mSettings.ucode=strtol(right(readinfo,1),NULL,10);
				
				if (lstrcmpi(left(readinfo,7), "Comment")==0)
					strcpy(mSections[sectionno].mSettings.szComment,right(readinfo,strlen(readinfo)-8));
				
				if (lstrcmpi(left(readinfo,4), "Info")==0)
					strcpy(mSections[sectionno].mSettings.szInfo,right(readinfo,strlen(readinfo)-5));
				
				if (lstrcmpi(left(readinfo,4), "Name")==0)
					strcpy(mSections[sectionno].mSettings.szGameName,right(readinfo,strlen(readinfo)-5));
				
				if (lstrcmpi(left(readinfo,14), "DisableDynarec")==0)
					mSections[sectionno].mSettings.bDisableDynarec=true;
				
				if (lstrcmpi(left(readinfo,14), "DisablePatches")==0)
					mSections[sectionno].mSettings.bDisablePatches=true;
				
				if (lstrcmpi(left(readinfo,20), "DisableTextureCRC")==0)
					mSections[sectionno].mSettings.bDisableTextureCRC=true;
				
				if (lstrcmpi(left(readinfo,13), "DisableEeprom")==0)
					mSections[sectionno].mSettings.bDisableEeprom=true;
				
				if (lstrcmpi(left(readinfo,16), "DisableSpeedSync")==0)
					mSections[sectionno].mSettings.bDisableSpeedSync=true;
				
				if (lstrcmpi(left(readinfo,14), "IncTexRectEdge")==0)
					mSections[sectionno].mSettings.bIncTexRectEdge=true;

				if (lstrcmpi(left(readinfo,12), "ExpansionPak")==0)
					mSections[sectionno].mSettings.bExpansionPak = true;

				if (lstrcmpi(left(readinfo,10), "EepromSize")==0)
					mSections[sectionno].mSettings.dwEepromSize = strtol(right(readinfo,4),NULL,10);

				if (lstrcmpi(left(readinfo,13), "OSRescanCount")==0)
					mSections[sectionno].mSettings.dwRescanCount = strtol(right(readinfo,4),NULL,10);
				
			}
		}
	}
	inifile.close();

	return true;
}

//*****************************************************************************
//read a line from the ini file
//*****************************************************************************
std::istream & IIniFile::getline(std::istream & is, char *str)
{
    char buf[100];
	
    is.getline(buf,100);
    strcpy( str,buf);
    return is;
}

//*****************************************************************************
//
//*****************************************************************************
void IIniFile::Commit()
{
	TCHAR szFileNameOut[MAX_PATH+1];
	TCHAR szFileNameDelete[MAX_PATH+1];
	u32 i;
	FILE * fhIn;
	FILE * fhOut;
	TCHAR szBuf[1024+1];
	char trim[]="{}\n\r"; //remove first and last character

	wsprintf(szFileNameOut, "%s.tmp", mFileName);
	wsprintf(szFileNameDelete, "%s.del", mFileName);

	fhIn = fopen(mFileName, "r");
	if (fhIn == NULL)
		return;

	fhOut = fopen(szFileNameOut, "w");
	if (fhOut == NULL)
	{
		fclose(fhIn);
		return;
	}

	// Mark all mSections and needing to be written
	for (i = 0; i < mSections.size(); i++)
	{
		mSections[i].bOutput = false;
	}


	while (fgets(szBuf, 1024, fhIn))
	{
		if (szBuf[0] == '{')
		{
			BOOL bFound = FALSE;
			// Start of section
			StrTrim(szBuf,trim);

			RomID id = RomIDFromString( szBuf );

			for (i = 0; i < mSections.size(); i++)
			{
				if (mSections[i].bOutput)
					continue;

				if ( mSections[i].mRomID == id )
				{
					// Output this CRC
					OutputSectionDetails(i, fhOut);
					mSections[i].bOutput = true;
					bFound = TRUE;
					break;
				}
			}
			if (!bFound)
			{
				// Do what? This should never happen, unless the user
				// replaces the inifile while Daedalus is running!
			}
		}
		else if (szBuf[0] == '/')
		{
			// Comment
			fputs(szBuf, fhOut);
			continue;
		}

	}

	// Input buffer done-  process any new entries!
	for (i = 0; i < mSections.size(); i++)
	{
		// Skip any that have not been done.
		if (mSections[i].bOutput)
			continue;
		// Output this CRC
		// Removed at request of Genueix :)
		//fprintf(fhOut, "// Automatically generated entry - may need editing\n");
		OutputSectionDetails(i, fhOut);
		mSections[i].bOutput = true;
	}

	fclose(fhOut);
	fclose(fhIn);

	// Create the new file
	MoveFile(mFileName, szFileNameDelete);
	MoveFile(szFileNameOut, mFileName);
	DeleteFile(szFileNameDelete);

	mDirty = false;
}


//*****************************************************************************
//
//*****************************************************************************
void IIniFile::OutputSectionDetails(DWORD i, FILE * fh)
{
	// Generate the CRC-ID for this rom:
	fprintf(fh, "{%08x%08x-%02x}\n", mSections[i].mRomID.CRC[0], mSections[i].mRomID.CRC[1], mSections[i].mRomID.CountryID );
	
	fprintf(fh, "Name=%s\n", mSections[i].mSettings.szGameName);
	fprintf(fh, "UCode=%d\n", mSections[i].mSettings.ucode);

	if (lstrlen(mSections[i].mSettings.szComment) > 0)	fprintf(fh, "Comment=%s\n", mSections[i].mSettings.szComment);
	if (lstrlen(mSections[i].mSettings.szInfo) > 0)		fprintf(fh, "Info=%s\n", mSections[i].mSettings.szInfo);
	if (mSections[i].mSettings.bDisableDynarec)			fprintf(fh, "DisableDynarec\n");
	if (mSections[i].mSettings.bDisablePatches)			fprintf(fh, "DisablePatches\n");
	if (mSections[i].mSettings.bDisableTextureCRC)		fprintf(fh, "DisableTextureCRC\n");
	if (mSections[i].mSettings.bDisableEeprom)			fprintf(fh, "DisableEeprom\n");
	if (mSections[i].mSettings.bDisableSpeedSync)			fprintf(fh, "DisableSpeedSync\n");
	if (mSections[i].mSettings.bIncTexRectEdge)			fprintf(fh, "IncTexRectEdge\n");
	if (mSections[i].mSettings.bExpansionPak)				fprintf(fh, "ExpansionPak\n");
	if (mSections[i].mSettings.dwEepromSize != 2048)		fprintf(fh, "EepromSize=%d\n", mSections[i].mSettings.dwEepromSize);
	if (mSections[i].mSettings.dwRescanCount != 0)		fprintf(fh, "OSRescanCount=%d\n", mSections[i].mSettings.dwRescanCount);

	fprintf(fh, "\n");			// Spacer
}

//*****************************************************************************
// Internal function to add a new entry to the database
//*****************************************************************************
void IIniFile::AddEntry( const RomID & id, const RomSettings & settings )
{
	// Add new entry!!!
	section newsection;

	newsection.mRomID = id;
	newsection.bOutput = false;
	newsection.mSettings = settings;

	mSections.push_back(newsection);

	mDirty = true;				// Flag to indicate we should be updated
}

//*****************************************************************************
// Internal function to find the entry corresponding to the specified rom. 
//*****************************************************************************
int IIniFile::FindEntry( const RomID & id ) const
{
	for ( u32 i = 0; i < mSections.size(); i++)
	{
		if ( mSections[i].mRomID == id )
			return i;
	}

	//
	// Not found
	//
	return -1;
}

//*****************************************************************************
// Retreive the settings for the specified rom. Returns false if the rom is 
// not in the database
//*****************************************************************************
bool	IIniFile::GetSettings( const RomID & id, RomSettings * p_settings ) const
{
	int i = FindEntry( id );
	if ( i >= 0 )
	{
		*p_settings = mSections[ i ].mSettings;
		return true;
	}
	else
	{
		return false;
	}
}

//*****************************************************************************
// Update the settings for the specified rom - creates a new entry if necessary
//*****************************************************************************
void	IIniFile::SetSettings( const RomID & id, const RomSettings & settings )
{
	int i = FindEntry( id );
	if ( i >= 0 )
	{
		mSections[ i ].mSettings = settings;
	}
	else
	{
		AddEntry( id, settings );
	}
	
	mDirty = true;
}

